/*
 * File   :     PbnGib.java
 * Author :     Tis Veugen
 * Date   :     2005-01-22
 * PBN    :     2.0
 *
 * History
 * -------
 * 2005-01-22 Adapted for GIB Version 6.1.3
 */

import java.io.*;

public class PbnGib
{
  private GibProcess            mGibProcess;
  private int                   mNrTricks;

  private static final String   GIB_FILENAME = "pbnJgib.dd";

  public			PbnGib()
  {
    mGibProcess = new GibProcess();
  }

  public PbnHand                GetOptCards(
  PbnDeal                       oDeal,
  PbnTrump                      oTrump,
  PbnPlayAdmin                  oPlayAdmin,
  int                           iCardIndex )
  {
    GibFile                     lGibFile;

    try
    {
      lGibFile = new GibFile( GIB_FILENAME );
    }
    catch ( Exception e )
    {
      return null;
    }

    PbnDeal                     lDeal = new PbnDeal( oDeal );
    int                         iNrPlayed = iCardIndex & 0x03;
    PbnSide                     lLeadSide;
    PbnSide                     lPlaySide;
    PbnHand                     lPlayHand;

    /*
     * Compute the playing side when no cards have been played yet
     * in the current trick.
     */
    PbnPlayCard lPlayCard = oPlayAdmin.GetPlayCard( iCardIndex );
    lPlaySide = new PbnSide( lPlayCard.GetSide() );
    lPlayHand = lDeal.GetHand( lPlaySide );
    lLeadSide = new PbnSide( lPlaySide );

    /*
     * Unplay the played cards.
     */
    for ( int i = 0; i < iNrPlayed; i++ )
    {
      lPlayCard = oPlayAdmin.GetPlayCard( iCardIndex-1 - i );
      lLeadSide = lPlayCard.GetSide();
      lDeal.UnplayCard( lLeadSide
                      , lPlayCard.GetCard() );
    }

    try
    {
      lGibFile.WriteDeal( lDeal, lPlaySide );
      lGibFile.WriteLeaderAndTrump( iNrPlayed, oTrump );
      lGibFile.WritePlayedCards( oPlayAdmin, iCardIndex );
      lGibFile.Close();
    }
    catch ( IOException e )
    {
      return null;
    }

    String [] laOptCards = mGibProcess.GetOptCards( GIB_FILENAME );
    if ( laOptCards == null )
    {
      return null;
    }

    mNrTricks = mGibProcess.GetOptNrTricks();

    /*
     * Translate the string array to a card array.
     */
    int                         iNrCards = laOptCards.length;
    PbnHand                     lOptHand = new PbnHand();
    String                      lString;
    PbnHand                     lHand;
    PbnHand                     lOtherHands = new PbnHand();
    PbnSuit                     lSuit = new PbnSuit();
    PbnRank                     lRank;

    /*
     * Add the ranks of the 3 other hands.
     */
    for ( int iSide = 0; iSide < PbnSide.NUMBER-1; iSide++ )
    {
      lPlaySide.Next();
      lHand = lDeal.GetHand( lPlaySide );

      lSuit.Set( PbnSuit.CLUBS );
      for ( int iSuit = 0; iSuit < PbnSuit.NUMBER; iSuit++ )
      {
        lOtherHands.AddRanks( lSuit, lHand.GetRanks( lSuit ) );
        lSuit.Next();
      }
    }

    /*
     * Add a sentinel to the cards.
     */
    lRank = new PbnRank( PbnRank.MIN );
    lSuit.Set( PbnSuit.CLUBS );
    for ( int iSuit = 0; iSuit < PbnSuit.NUMBER; iSuit++ )
    {
      lOtherHands.AddRank( lSuit, lRank );
      lSuit.Next();
    }

    for ( int j = 0; j < iNrCards; j++ )
    {
      lString = laOptCards[j];
      {
        lSuit = new PbnSuit( PbnSuit.FromChar( lString.charAt( 0 ) ) );
        lRank = new PbnRank( PbnRank.FromChar( lString.charAt( 1 ) ) );

        /*
         * Add the ranks in the hand <= 'rank', but > than a rank from another hand.
         */
        while ( ! lOtherHands.HasRank( lSuit, lRank ) )
        {
          if ( lPlayHand.HasRank( lSuit, lRank ) )
          {
            lOptHand.AddRank( lSuit, lRank );
          }
          lRank.Previous();
        }
      }
    }

    return lOptHand;
  }

  public int                    GetOptNrTricks()
  {
    return mNrTricks;
  }

  public void                   Quit()
  {
    mGibProcess.Destroy();
  }
}

class GibFile
{
  DataOutputStream              mGibOs;

  public                        GibFile(
  String                        oFilename )
  {
    try
    {
      FileOutputStream            lFos = new FileOutputStream( oFilename );
      BufferedOutputStream        lBos = new BufferedOutputStream( lFos );
      DataOutputStream            lDos = new DataOutputStream( lBos );
      mGibOs = lDos;
    }
    catch ( Exception e )
    {
      mGibOs = null;
    }
  }

  private void                  Write(
  StringBuffer                  oSb )
  throws IOException
  {
    oSb.append( "\n" );
    mGibOs.writeBytes( oSb.toString() );
  }

  public void                   WriteDeal(
  PbnDeal                       oDeal,
  PbnSide                       lPlaySide )
  throws IOException
  {
    PbnSide                     lSide;
    PbnSuit                     lSuit = new PbnSuit();
    StringBuffer                lSb;
    String                      lString;

    lSide = new PbnSide( PbnSide.WEST );
    for ( int iSide = 0; iSide < PbnSide.NUMBER; iSide++ )
    {
      lPlaySide.Next();

      lSb = new StringBuffer( lSide.toCharacter().toLowerCase() );
      lSb.append( " " );
      lSuit.Set( PbnSuit.SPADES );
      for ( int iSuit = 0; iSuit < PbnSuit.NUMBER; iSuit++ )
      {
        lString = oDeal.GetRanks( lPlaySide, lSuit ).toString();
        lSb.append( lString );
        if ( iSuit < PbnSuit.NUMBER-1 )
        {
          lSb.append( "." );
        }
        lSuit.Previous();
      }
      Write( lSb );
      lSide.Next();
    }
  }

  public void                   WriteLeaderAndTrump(
  int                           iNrPlayed,
  PbnTrump                      oTrump )
  throws IOException
  {
    PbnSide                     lLeadSide = new PbnSide( PbnSide.SOUTH );

    for ( int i = 0; i < iNrPlayed; i++ )
    {
      lLeadSide.Previous();
    }

    StringBuffer                lSb;

    lSb = new StringBuffer( lLeadSide.toCharacter().toLowerCase() );
    lSb.append( " " + oTrump.toString().toLowerCase() );
    /*
     * Possibly remove the 't' of 'nt'.
     */
    lSb.setLength( 3 );

    Write( lSb );
  }

  public void                   WritePlayedCards(
  PbnPlayAdmin                  oPlayAdmin,
  int                           iCardIndex )
  throws IOException
  {
    StringBuffer                lSb = new StringBuffer();
    PbnPlayCard                 lPlayCard;
    int                         iNrPlayed = iCardIndex & 0x03;

    for ( int i = 0; i < iNrPlayed; i++ )
    {
      lPlayCard = oPlayAdmin.GetPlayCard( iCardIndex-iNrPlayed + i );
      lSb.append( lPlayCard.GetCard().toString().toLowerCase() );
      lSb.append( " " );
    }

    Write( lSb );
  }

  public void                   Close()
  throws IOException
  {
    mGibOs.flush();
    mGibOs.close();
  }
}
